    -- ResolverAD2S1200DecA7  - Resolver connected to AD2S1200 decoder with CmodA7 board
    -- Copyright (C) 2021 VCC
    -- release date: 31 Oct 2021
    -- author: VCC

    -- This file is part of ResolverAD2S1200DecA7 project.

    -- Permission is hereby granted, free of charge, to any person obtaining a copy
    -- of this software and associated documentation files (the "Software"),
    -- to deal in the Software without restriction, including without limitation
    -- the rights to use, copy, modify, merge, publish, distribute, sublicense,
    -- and/or sell copies of the Software, and to permit persons to whom the
    -- Software is furnished to do so, subject to the following conditions:

    -- The above copyright notice and this permission notice shall be included
    -- in all copies or substantial portions of the Software.

    -- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    -- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    -- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    -- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    -- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    -- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
    -- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


----------------------------------------------------------------------------------
-- Company: 
-- Engineer: VCC
-- 
-- Create Date:    11:15:56 10/17/2021 
-- Design Name: 
-- Module Name:    AD2SState - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: state machine for controlling AD2S startup/reset sequence and error signals
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity AD2SState is
    Port (Clk1MHz: in STD_LOGIC;                --Clock source for this component. Must be 1MHz, as timings are derived from it.
          Reset: in STD_LOGIC;                  --Reset signal for this component, active low. Can be tied to main "Locked" signal of the design.
          LOT: in STD_LOGIC;                    --Loss of track (output from AD2S1200), active low
          DOS: in STD_LOGIC;                    --Degradation of signal (output from AD2S1200), active low
          FilteredLOT: out STD_LOGIC;           --Persisted state of LOT, active high as a status indicator
          FilteredDOS: out STD_LOGIC;           --Persisted state of DOS, active high as a status indicator
          LOS: out STD_LOGIC;                   --Loss of signal (combination of LOT and DOS, when both are active), active high as a status indicator
          AD2S_Reset: out STD_LOGIC;            --AD2S1200 Reset signal, active low
          Allow_AD2S_Reset_OnLOS: in STD_LOGIC; --Allows the state machine to reset the AD2S chip, active 1
          AD2S_Reset_FromUI: in STD_LOGIC;      --Manual reset from UI, active low
          ValidAfterReset: out STD_LOGIC        --Transitions to 1, when the next SAMPLE signal is allowed to go to 0  (see AD2S1200 startup sequence)
          );
end AD2SState;


architecture Behavioral of AD2SState is
 
  component AD2SErrorHandling is
    Port (FilterClk: in STD_LOGIC;   --Clock source for this component. Must be 1MHz / 1024 = 976.5625Hz, as timings are derived from it.
          SyncLOT: in STD_LOGIC;     --Synchronized version of LOT signal - Loss of track (output from AD2S1200), active low
          SyncDOS: in STD_LOGIC;     --Synchronized version of DOS signal - Degradation of signal (output from AD2S1200), active low
          FilteredLOT: out STD_LOGIC;--Persisted state of LOT, active high as a status indicator
          FilteredDOS: out STD_LOGIC;--Persisted state of DOS, active high as a status indicator
          LOS: out STD_LOGIC         --Loss of signal (combination of LOT and DOS, when both are active), active high as a status indicator
    );
  end component;

  type TStates is (SInit, SIdle, SResetAD2S, SReleaseResetAD2S, SStartTTrack, SEndTTrack, SWaitAfterReset, SWaitForCounterReset);
  signal State: TStates := SInit;
  signal NextState: TStates := SIdle;
  
  constant CLOSCounterMax: STD_LOGIC_VECTOR(9 downto 0) := CONV_STD_LOGIC_VECTOR(1022, 10); --controls when to stop accumulating the LOS counter, while LOS is active  (1.024[ms] * 1022 = 1.046528[s])
  constant CLOSCounterFSMMax: STD_LOGIC_VECTOR(9 downto 0) := CONV_STD_LOGIC_VECTOR(1000, 10); --controls how much to wait for LOS signal to be in the active state, in order to trigger AD2S Resetting  (1.024[ms] * 1000 = 1.024[s]). It has to be less than CLOSCounterMax.
  
  constant CWaitingAfterResetCounterMax: STD_LOGIC_VECTOR(10 downto 0) := CONV_STD_LOGIC_VECTOR(2022, 11); --controls when to stop accumulating the WaitAfterResetCounter counter, while EnWaitingAfterReset is active  (1.024[ms] * 2022 = 2.070528[s])
  constant CWaitingAfterResetCounterFSMMax: STD_LOGIC_VECTOR(10 downto 0) := CONV_STD_LOGIC_VECTOR(2000, 11); --controls when to stop accumulating the WaitAfterResetCounter counter, while in SWaitAfterReset state  (1.024[ms] * 2000 = 2.048[s]).  Must be less than CWaitingAfterResetCounterMax.
  
  constant CResetAD2SCounterFSMMax: STD_LOGIC_VECTOR(15 downto 0) := CONV_STD_LOGIC_VECTOR(1000, 16); --controls then length of AD2S RESET impulse in [us], e.g. 1ms
  constant CValidAfterResetAD2SCounterFSMMax: STD_LOGIC_VECTOR(15 downto 0) := CONV_STD_LOGIC_VECTOR(30000, 16); --controls then length of AD2S SAMPLE impulse in [us], e.g. 30ms

  signal MainCounter: STD_LOGIC_VECTOR(9 downto 0) := (others => '0'); --local clock divider
  signal FilterClk: STD_LOGIC;
  signal TempLOS: STD_LOGIC;
  signal SyncLOTReg: STD_LOGIC_VECTOR(2 downto 0) := (others => '0'); -- 3-bit shift register, to cancel metastability
  signal SyncDOSReg: STD_LOGIC_VECTOR(2 downto 0) := (others => '0'); -- 3-bit shift register, to cancel metastability
  signal SyncLOT: STD_LOGIC; --shift register output
  signal SyncDOS: STD_LOGIC; --shift register output
  
  signal EnWaitingAfterReset: STD_LOGIC := '0'; --enables waiting a specified amount of time, after exiting from AD2S reset state
  
  signal LOS_Counter: STD_LOGIC_VECTOR(9 downto 0) := (others => '0');
  signal ResetAD2SCounter: STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
  signal WaitAfterResetCounter: STD_LOGIC_VECTOR(10 downto 0) := (others => '0'); --counts while EnWaitingAfterReset = '1'
 
begin

  process(Clk1MHz)
  begin
    if Clk1MHz'event and Clk1MHz = '1' then
      if Reset = '0' then
        MainCounter <= (others => '0');
      else
        MainCounter <= MainCounter + 1;        
        SyncLOTReg <= SyncLOTReg(1 downto 0) & LOT;
        SyncDOSReg <= SyncDOSReg(1 downto 0) & DOS;
      end if;
    end if;
  end process;
  
  SyncLOT <= SyncLOTReg(2);
  SyncDOS <= SyncDOSReg(2);
  
  FilterClk <= MainCounter(9); --bit 9 is 1MHz / 1024 = 976.5625Hz
  
  AD2SErrorHandling_Instance: AD2SErrorHandling port map (
    FilterClk => FilterClk,
    SyncLOT => SyncLOT,
    SyncDOS => SyncDOS,
    FilteredLOT => FilteredLOT,
    FilteredDOS => FilteredDOS,
    LOS => TempLOS
  );
  
  LOS <= TempLOS;
  
    
  process(FilterClk)
  begin
    if FilterClk'event and FilterClk = '1' then  --976.5625Hz
      if TempLOS = '1' then
        if LOS_Counter <= CLOSCounterMax then
          LOS_Counter <= LOS_Counter + 1;
        end if;
      else
        if State = SWaitForCounterReset then --use a special state for this, because states are driven by another clock
          LOS_Counter <= (others => '0');
        end if;
      end if;
            
      if EnWaitingAfterReset = '1' then
        if WaitAfterResetCounter < CWaitingAfterResetCounterMax then
          WaitAfterResetCounter <= WaitAfterResetCounter + 1;
        end if;
      else
        WaitAfterResetCounter <= (others => '0');
      end if;
    end if;
  end process;

  
  -- Startup FSM:
  
  process(Clk1MHz)
  begin
    if Clk1MHz'event and Clk1MHz = '1' then
      if Reset = '0' then
        State <= SInit;
      else
        State <= NextState;
      end if;
    end if;
  end process;
  
   
  process(State, Allow_AD2S_Reset_OnLOS, ResetAD2SCounter, LOS_Counter, WaitAfterResetCounter, AD2S_Reset_FromUI)
  begin
    case State is
      when SInit =>
        NextState <= SIdle;
        
      when SIdle =>
        if (LOS_Counter >= CLOSCounterFSMMax) or (AD2S_Reset_FromUI = '1') then
          NextState <= SResetAD2S;
        else
          NextState <= SIdle;
        end if;
        
      when SResetAD2S =>
        if ResetAD2SCounter <= CResetAD2SCounterFSMMax then
          NextState <= SResetAD2S;
        else
          NextState <= SReleaseResetAD2S;
        end if;
        
      when SReleaseResetAD2S =>
        NextState <= SStartTTrack;
        
      when SStartTTrack =>
        if ResetAD2SCounter <= CValidAfterResetAD2SCounterFSMMax then
          NextState <= SStartTTrack;
        else
          NextState <= SEndTTrack;
        end if;

      when SEndTTrack =>
        if (Allow_AD2S_Reset_OnLOS = '1') and (WaitAfterResetCounter = 0) then
          NextState <= SWaitAfterReset;
        else
          NextState <= SEndTTrack;
        end if;
        
      when SWaitAfterReset =>
        if WaitAfterResetCounter <= CWaitingAfterResetCounterFSMMax then
          NextState <= SWaitAfterReset;
        else
          NextState <= SWaitForCounterReset;
        end if;
        
      when SWaitForCounterReset =>
        if LOS_Counter = 0 then
          NextState <= SIdle;
        else
          NextState <= SWaitForCounterReset;
        end if;
        
      when others =>
        NextState <= SIdle;
    end case;
  end process;
  
  
  process(Clk1MHz)
  begin
    if Clk1MHz'event and Clk1MHz = '0' then
      case State is
        when SInit =>
          ValidAfterReset <= '0';
          
        when SIdle =>
          ResetAD2SCounter <= (others => '0');
          EnWaitingAfterReset <= '0';
                  
        when SResetAD2S =>
          AD2S_Reset <= '0';
          ResetAD2SCounter <= ResetAD2SCounter + 1;
          ValidAfterReset <= '0';
          
        when SReleaseResetAD2S =>
          AD2S_Reset <= '1';
          ResetAD2SCounter <= (others => '0');
          
        when SStartTTrack =>
          ResetAD2SCounter <= ResetAD2SCounter + 1;
          
        when SEndTTrack =>
          ResetAD2SCounter <= (others => '0');
          ValidAfterReset <= '1';
          EnWaitingAfterReset <= '0';
          
        when SWaitAfterReset =>
          EnWaitingAfterReset <= '1';
          
        when SWaitForCounterReset =>
          
        when others =>
      end case;
    end if;
  end process;

end Behavioral;

